""" One particle in Morse potential - Chebyshev with adaptive time step """
import numpy as np
import matplotlib.pyplot as plt
from rotagaporp.chebyshev import Chebyshev, SymmetricChebyshev
from rotagaporp.systems import Morse
from utilities.timing import timeit

params = {'tstart': 0., 'tend': 100., 'tstep': 0.2, 'nterms': 4, 'tqdm': True,
          'varstep': True, 'tolerance': 1e-4}

syspar = {'q0': 3., 'p0': 0.}

system = Morse(**syspar)

# non-symmetrized propagator
ch_prop = Chebyshev(syst=system, **params)
print('Chebyshev propagator')
timeit(ch_prop.propagate)()
time_ch, q_ch, p_ch = ch_prop.get_trajectory()
ch_prop.analyse()
print('energy drift:', np.max(np.fabs(ch_prop.er)))
print('number of time steps:', len(ch_prop.tsteps))
print('min time step size:', min(ch_prop.tsteps))
print('max time step size:', max(ch_prop.tsteps))

# symmetrized propagator
sy_prop = SymmetricChebyshev(syst=system, **params)
print('\nSymmetric Chebyshev propagator')
timeit(sy_prop.propagate)()
time_sy, q_sy, p_sy = sy_prop.get_trajectory()
sy_prop.analyse()
print('energy drift:', np.max(np.fabs(sy_prop.er)))
print('number of time steps:', len(sy_prop.tsteps))
print('min time step size:', min(sy_prop.tsteps))
print('max time step size:', max(sy_prop.tsteps))

q_an, p_an = system.solution(times=time_sy)
en0 = system.energy(system.q0, system.p0)
evecf = np.vectorize(system.energy)
err_an = np.array((evecf(q_an, p_an)-en0)/en0, dtype=float)
print('\nAnalytic solution')
print('energy drift:', np.max(np.fabs(err_an)))

fig = plt.figure()
plot1 = fig.add_subplot(311)
plot1.plot(time_ch, np.maximum.accumulate(abs(ch_prop.er)), label='Chebyshev')
plot1.plot(time_sy, np.maximum.accumulate(abs(sy_prop.er)), label='Symmetric Chebyshev')
plot1.set_ylabel('maximum energy drift')
plot1.set_yscale('log')
plot1.set_xscale('log')
plot1.legend()

plot2 = fig.add_subplot(312)
plot2.plot(time_ch, q_ch, label='Chebyshev')
plot2.plot(time_sy, q_sy, label='Symmetric Chebyshev')
plot2.plot(time_sy, q_an, label='Analytic solution')
plot2.set_xlabel(r'$t$')
plot2.set_ylabel(r'$q(t)$')
plot2.set_xscale('log')
plot2.legend()

plot3 = fig.add_subplot(313)
plot3.plot(time_ch, p_ch, label='Chebyshev')
plot3.plot(time_sy, p_sy, label='Symmetric Chebyshev')
plot3.plot(time_sy, p_an, label='Analytic solution')
plot3.set_xlabel(r'$t$')
plot3.set_ylabel(r'$p(t)$')
plot3.set_xscale('log')
plot3.legend()

plt.show()
